package com.example.facetter;

import java.io.IOException;

import android.app.Activity;
import android.app.AlertDialog;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.PointF;
import android.graphics.RectF;
import android.hardware.Camera;
import android.media.FaceDetector;
import android.media.FaceDetector.Face;
import android.os.Bundle;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.Window;
import android.view.WindowManager;
import android.widget.FrameLayout;
import android.widget.ImageView;

public class CameraActivity extends Activity {

    private FrameLayout mFrameLayout; // ビューを重ねて表示するためのレイアウト
    private SurfaceView mCameraView;  // カメラプレビューのためのSurfaceView
    private FaceView mFaceView;       // プレビュー時にアイコン画像を描画するビュー
    private Camera mCamera;			  // カメラ

    private ImageView mImageView;	  // mPictureを表示するためのImageView
    private Bitmap mIcon;			  // アイコン画像
    private Bitmap mPicture;		  // 撮影、顔認識、アイコン画像を合成した画像
    private boolean mAutoFocusing = false; // オートフォーカス中の場合はtrue

    private static final int MENU_SAVE 	  = (Menu.FIRST + 1);
    private static final int MENU_DISCARD = (Menu.FIRST + 2);

    private SurfaceHolder.Callback mSurfaceHolderCallback = new SurfaceHolder.Callback() {
        public void surfaceCreated(SurfaceHolder holder) {
            // カメラを起動（オープン）します
            mCamera = Camera.open();
            try {
                mCamera.setPreviewDisplay(holder);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        public void surfaceChanged(SurfaceHolder holder, int format, int width,
                int height) {
            // カメラのプレビューのサイズを設定します
            Camera.Parameters params = mCamera.getParameters();
            params.setPreviewSize(width, height);
            mCamera.setParameters(params);
            // カメラのプレビューを開始します
            mCamera.setPreviewCallback(previewCallback);
            mCamera.startPreview();
        }

        public void surfaceDestroyed(SurfaceHolder holder) {
            // カメラのプレビューを停止します
            mCamera.setPreviewCallback(null);
            mCamera.stopPreview();
            // カメラを停止（解放）します
            mCamera.release();
            mCamera = null;
        }
    };

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        // フルスクリーン表示にします
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN);
        // タイトルバーを非表示にします
        requestWindowFeature(Window.FEATURE_NO_TITLE);

        // インテントで渡された値を受け取る
        mIcon = (Bitmap)getIntent().getExtras().getParcelable("icon");
        mFaceView = new FaceView(this);
        mFaceView.setIcon(mIcon);

        mFrameLayout = new FrameLayout(this);
        mCameraView = new SurfaceView(this);

        SurfaceHolder holder = mCameraView.getHolder();
        holder.addCallback(mSurfaceHolderCallback);
        holder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);

        mFrameLayout.addView(mCameraView);
        mFrameLayout.addView(mFaceView);
        setContentView(mFrameLayout);

        mImageView = new ImageView(this);
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // メニューアイテムを追加します
        menu.add(Menu.NONE, MENU_SAVE, Menu.NONE, "Save")
        .setIcon(android.R.drawable.ic_menu_save);
        menu.add(Menu.NONE, MENU_DISCARD, Menu.NONE, "Discard")
        .setIcon(android.R.drawable.ic_menu_delete);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        String filePath = null;
        switch (item.getItemId()) {
        // 撮影した画像を保存します。
        case MENU_SAVE:
//            result = savePicture();
            filePath = ImageManager.addImageAsApplication(this, mPicture);
            if (filePath != null) {
                Log.d(getClass().getName(), "saved image file path=" + filePath);
                // 成功
                showAlertDialog("Save picture", "Success!");
            } else {
                // 失敗
                showAlertDialog("Save picture", "Error! Sorry...");
            }
            return true;
        // 撮影した画像を破棄してカメラプレビューを再開します。
        case MENU_DISCARD:
            mPicture = null;
            mFrameLayout.removeView(mImageView);
            mCamera.startPreview();
            return true;
        }
        return true;
    }

    // 撮影した画像を保存します。
    private String savePicture() {
        String filename = "Facetter_" + System.currentTimeMillis();
        return android.provider.MediaStore.Images.Media.insertImage(
            getContentResolver(),
            mPicture,
            filename,
            null);
    }

    // 指定されたタイトルとメッセージを使ってアラートダイアログを表示します。
    private void showAlertDialog(String title, String message) {
        new AlertDialog.Builder(this)
        .setTitle(title)
        .setMessage(message)
        .setPositiveButton("OK", null)
        .create()
        .show();
    }

    private Camera.PreviewCallback previewCallback = new Camera.PreviewCallback() {
        public void onPreviewFrame(byte[] data, Camera camera) {
            // 二重起動しないようにPreviewCallbackをnullクリアしておく
            camera.setPreviewCallback(null);
            // プレビューのサイズを取得
            Camera.Size size = camera.getParameters().getPreviewSize();
            // プレビュー画像をYUV420SP形式からRGB形式に変換
            int rgb[] = new int[size.width * size.height];
            decodeYUV420SP(rgb, data, size.width, size.height);
            // Bitmap bitmap = Bitmap.createBitmap(size.width, size.height,
            // Bitmap.Config.ARGB_8888); // これじゃだめ
            // RGB形式の配列からBitmapを作る
            Bitmap bitmap = Bitmap.createBitmap(size.width, size.height,
                    Bitmap.Config.RGB_565);
            bitmap.setPixels(rgb, 0, size.width, 0, 0, size.width, size.height);

            FaceDetector.Face[] faces = new FaceDetector.Face[1];
            FaceDetector faceDetector = new FaceDetector(bitmap.getWidth(),
                    bitmap.getHeight(), faces.length);
            int num = faceDetector.findFaces(bitmap, faces);
            if (num > 0) {
                mFaceView.setFaces(faces);
            } else {
                mFaceView.setFaces(null);
            }
            // 描画を要求する
            mFaceView.postInvalidate();
            // 最初にPreviewCallbackをnullクリアしたので忘れずに再セットしておく
            camera.setPreviewCallback(previewCallback);
        }

        // YUV420SPからRGBに変換します。
        private void decodeYUV420SP(int[] rgb, byte[] yuv420sp, int width,
                int height) {
            final int frameSize = width * height;

            for (int j = 0, yp = 0; j < height; j++) {
                int uvp = frameSize + (j >> 1) * width, u = 0, v = 0;
                for (int i = 0; i < width; i++, yp++) {
                    int y = (0xff & ((int) yuv420sp[yp])) - 16;
                    if (y < 0)
                        y = 0;
                    if ((i & 1) == 0) {
                        v = (0xff & yuv420sp[uvp++]) - 128;
                        u = (0xff & yuv420sp[uvp++]) - 128;
                    }

                    int y1192 = 1192 * y;
                    int r = (y1192 + 1634 * v);
                    int g = (y1192 - 833 * v - 400 * u);
                    int b = (y1192 + 2066 * u);

                    if (r < 0)
                        r = 0;
                    else if (r > 262143)
                        r = 262143;
                    if (g < 0)
                        g = 0;
                    else if (g > 262143)
                        g = 262143;
                    if (b < 0)
                        b = 0;
                    else if (b > 262143)
                        b = 262143;

                    rgb[yp] = 0xff000000 | ((r << 6) & 0xff0000)
                            | ((g >> 2) & 0xff00) | ((b >> 10) & 0xff);
                }
            }
        }
    };

    private Camera.AutoFocusCallback autoFocusCallback = new Camera.AutoFocusCallback() {
        public void onAutoFocus(boolean success, Camera camera) {
            if (success) {
                // 写真撮る
                camera.takePicture(null, null, pictureCallback);
            } else {
                camera.setPreviewCallback(previewCallback);
            }
            mAutoFocusing = false;
        }
    };

    private final Camera.PictureCallback pictureCallback = new Camera.PictureCallback() {
        public void onPictureTaken(byte[] data, Camera camera) {
            Bitmap maskFaceImage;
            BitmapFactory.Options options = new BitmapFactory.Options();
            // 調べるだけフラグをON!
            options.inJustDecodeBounds = true;
            // サイズを調べるだけ
            BitmapFactory.decodeByteArray(data, 0, data.length, options);
            // scaleを計算
            int scaleW = options.outWidth / 720 + 1;
            int scaleH = options.outHeight / 480 + 1;
            int scale = Math.max(scaleW, scaleH);
            // 調べるだけフラグを下ろして、inSampleSizeをセット
            options.inJustDecodeBounds = false;
            options.inSampleSize = scale;
            Bitmap picture = BitmapFactory.decodeByteArray(data, 0,
                    data.length, options);
            Bitmap faceBitmap = picture.copy(Bitmap.Config.RGB_565, true);

            Log.d(getClass().getName(), "taken picture size width="
                    + picture.getWidth() + ", height=" + picture.getHeight());

            FaceDetector.Face[] faces = new FaceDetector.Face[1];
            FaceDetector faceDetector = new FaceDetector(faceBitmap.getWidth(),
                    faceBitmap.getHeight(), faces.length);
            int num = faceDetector.findFaces(faceBitmap, faces);

            if (num > 0) {
                // 顔認識結果が1つ以上あれば画像を合成
                Bitmap offScreenBitmap = Bitmap.createBitmap(
                        picture.getWidth(), picture.getHeight(),
                        Bitmap.Config.ARGB_8888);
                Canvas offScreen = new Canvas(offScreenBitmap);
                offScreen.drawBitmap(picture, 0, 0, null);

                for (int i = 0; i < faces.length; i++) {
                    Face face = faces[i];
                    if (face == null)
                        break;
                    PointF point = new PointF(0, 0);
                    face.getMidPoint(point);
                    float eyesDistance = face.eyesDistance();
                    RectF rect = new RectF();
                    rect.left = point.x - eyesDistance * 2;
                    rect.top = point.y - eyesDistance * 2;
                    rect.right = point.x + eyesDistance * 2;
                    rect.bottom = point.y + eyesDistance * 2;
                    offScreen.drawBitmap(mIcon, null, rect, null);
                }

                maskFaceImage = offScreenBitmap;
            } else {
                // 認識結果がない場合は、撮影した画像をそのまま
                maskFaceImage = picture;
            }
            mPicture = maskFaceImage;
            mImageView.setImageBitmap(maskFaceImage);
            mImageView.setBackgroundColor(Color.BLACK);
            mFrameLayout.addView(mImageView);
        }
    };

    @Override
    public boolean onTouchEvent(MotionEvent event) {
        // ACTION_DOWNだけ実行するように判定
        if (event.getAction() == MotionEvent.ACTION_DOWN) {
            if (mPicture == null) {
                // mPictureが存在しない時は、プレビュー状態なので
                if (mAutoFocusing == false) {
                    // オートフォーカス中でなければ、オートフォーカスを開始
                    mCamera.autoFocus(autoFocusCallback);
                    mAutoFocusing = true;
                }
            } else {
                // mPictureが存在する時は、画像を表示している状態なので
                // タッチされたらメニューを表示
                getWindow().openPanel(Window.FEATURE_OPTIONS_PANEL, null);
            }
        }
        return true;
    }
}
